@@ -75,6 +75,7 @@ evaluate (Val v@(MVal mv)) = do
         else do
             rv <- liftIO (readIORef mv)
             evaluate (Val rv)
+evaluate (Val (VThunk (MkThunk t))) = t
 evaluate (Val val) = do
     -- context casting, go!
     cxt <- asks envContext
@@ -98,7 +99,6 @@ evaluate exp = do
     case val of
         VError s e  -> retError s e
         _           -> return val
-    return val
 
 evalExp :: Exp -> Eval Val
 evalExp exp = do
@@ -106,8 +106,10 @@ evalExp exp = do
     evl exp
 
 evalSym :: Symbol -> Eval (String, Val)
-evalSym (Symbol _ name vexp) = do
-    val <- evalExp vexp
+evalSym (SymVal _ name val) =
+    return (name, val)
+evalSym (SymExp _ name vexp) = do
+    val <- enterEvalContext (cxtOfSigil $ head name) vexp
     return (name, val)
     
 enterEvalContext :: Cxt -> Exp -> Eval Val
@@ -142,39 +144,43 @@ reduceStatements [] = reduceExp
 reduceStatements ((exp, pos):rest)
     | Syn ";" exps <- exp = do
         reduceStatements $ (exps `zip` repeat pos) ++ rest
-    | Syn "sym" (Sym sym@(Symbol _ _ vexp@(Syn "sub" [_])):other) <- exp = \v -> do
-        (VSub sub) <- enterEvalContext "Code" vexp
-        lex <- asks envLexical
-        reduceStatements ((Syn "sym" (other ++ [Sym sym{ symExp = Val $ VSub sub{ subPad = lex } }]), pos):rest) v
-    | Syn "sym" (Sym sym@(Symbol _ name (Syn "mval" [_, vexp])):other) <- exp = \_ -> do
-        val <- enterEvalContext (cxtOfSigil $ head name) vexp
-        mval <- newMVal val
-        reduceStatements ((Syn "sym" (other ++ [Sym sym{ symExp = Val mval }]), pos):rest) $ Val mval
-    | Syn "sym" [Sym sym@(Symbol SGlobal _ vexp)] <- exp = \_ -> do
+    | Sym [] <- exp = \v -> do
+        reduceStatements rest v
+    | Sym ((SymExp scope name vexp):other) <- exp = \v -> do
+        val <- enterLValue $ enterEvalContext (cxtOfSigil $ head name) vexp
+        reduceStatements ((Sym (SymVal scope name val:other), pos):rest) v
+    | Sym (sym@(SymVal scope _ val):other) <- exp
+    , scope == SOur || scope == SGlobal = \_ -> do
         addGlobalSym sym
-        reduceStatements rest vexp
-    | Syn "sym" [Sym sym@(Symbol SOur _ vexp)] <- exp = \_ -> do
-        addGlobalSym sym -- XXX Wrong
-        reduceStatements rest vexp
-    | Syn "sym" syms <- exp = \_ -> do
-        enterLex [ sym | Sym sym@(Symbol SMy _ _) <- syms] $ do
-            reduceStatements rest $ Syn "sym" syms
+        reduceStatements ((Sym other, pos):rest) (Val val)
+    | Sym (sym@(SymVal SMy _ val):other) <- exp = \_ -> do
+        enterLex [ sym ] $ do
+            reduceStatements ((Sym other, pos):rest) (Val val)
     | Syn syn [Var name, vexp] <- exp
     , (syn == ":=" || syn == "::=") = \_ -> do
-        lex <- asks envLexical
+        env <- ask
+        let lex = envLexical env
+            val = VThunk . MkThunk $ do
+            local (const env{ envLValue = True }) $ do
+                enterEvalContext (cxtOfSigil $ head name) vexp
         case findSym name lex of
             Just _  -> do
-                let sym = (Symbol SMy name vexp)
+                let sym = (SymVal SMy name val)
                 enterLex [sym] $ do
-                    reduceStatements rest vexp
+                    reduceStatements rest (Val val)
             Nothing -> do
-                addGlobalSym $ Symbol SGlobal name vexp
-                reduceStatements rest vexp
+                addGlobalSym $ SymVal SGlobal name val
+                reduceStatements rest (Val val)
     | Syn "sub" [Val (VSub sub)] <- exp
     , subType sub >= SubBlock = do
         -- bare Block in statement level; run it!
         let app = Syn "()" [exp, Syn "invs" [], Syn "args" []]
         reduceStatements $ (app, pos):rest
+    | Syn "dump" [] <- exp
+    , null rest = \e -> do
+        Env{ envGlobal = globals, envLexical = lexicals } <- ask
+        liftIO $ modifyIORef globals (lexicals ++)
+        reduceStatements rest e
     | null rest = \_ -> do
         _   <- asks envContext
         val <- enterLex (posSyms pos) $ reduceExp exp
@@ -190,7 +196,7 @@ reduceStatements ((exp, pos):rest)
         VError str exp  -> retError str exp
         _               -> action
 
-posSyms pos = [ Symbol SMy n (Val v) | (n, v) <- syms ]
+posSyms pos = [ SymVal SMy n v | (n, v) <- syms ]
     where
     file = sourceName pos
     line = show $ sourceLine pos
@@ -201,53 +207,37 @@ posSyms pos = [ Symbol SMy n (Val v) | (n, v) <- syms ]
         ]
 
 evalVar name = do
-    _   <- ask
-    val <- enterLValue $ do
-        rv <- findVar name
-        enterEvalContext (cxtOfSigil $ head name) $ case rv of
-            Nothing -> (Val $ VError ("Undefined variable " ++ name) (Val VUndef))
-            Just (Val val)  -> (Val val)
-            Just exp        -> exp -- XXX Wrong
-    return val
+    env <- ask
+    v <- findVar env name
+    return $ case v of
+        Just val -> val
+        Nothing -> VError ("Undefined variable " ++ name) (Val VUndef)
 
 enterLValue = local (\e -> e{ envLValue = True })
 
-breakOnGlue _ [] = ([],[])
-breakOnGlue glue rest@(x:xs)
-    | glue `isPrefixOf` rest = ([], rest)
-    | otherwise = (x:piece, rest') where (piece, rest') = breakOnGlue glue xs
-
-findVar name
-    | (package, name') <- breakOnGlue "::" name
-    , (sig, "CALLER") <- breakOnGlue "CALLER" package = do
-        rv <- asks envCaller
-        case rv of
-            Just caller -> findVar' caller (sig ++ (drop 2 name'))
+findVar :: Env -> Ident -> Eval (Maybe Val)
+findVar env name
+    | Just (package, name') <- breakOnGlue "::" name
+    , Just (sig, "") <- breakOnGlue "CALLER" package =
+        case (envCaller env) of
+            Just caller -> findVar caller (sig ++ name')
             Nothing -> retError "cannot access CALLER:: in top level" (Var name)
     | otherwise = do
-        env <- ask
-        findVar' env name
-    where
-    findVar' env name = do
-        let lexSym = findSym name $ envLexical env
-        -- XXX rewrite using Maybe monad
-        if isJust lexSym then return lexSym else do
+        callCC $ \foundIt -> do
+            let lexSym = findSym name $ envLexical env
+            when (isJust lexSym) $ foundIt lexSym
             glob <- liftIO . readIORef $ envGlobal env
             let globSym = findSym name glob
-            if isJust globSym
-                then return globSym
-                else
-                    let globSym = findSym (toGlobal name) glob in
-                    if isJust globSym
-                        then return globSym
-                        else return Nothing
-    
+            when (isJust globSym) $ foundIt globSym
+            let globSym = findSym (toGlobal name) glob
+            when (isJust globSym) $ foundIt globSym
+            return Nothing
+
 reduce :: Env -> Exp -> Eval Val
 
 -- Reduction for mutables
-reduce _ (Val val@(MVal _)) = do
-    lvalue  <- asks envLValue
-    if lvalue
+reduce env (Val val@(MVal _)) = do
+    if envLValue env
         then retVal val
         else do
             rv <- readMVal val
@@ -255,14 +245,13 @@ reduce _ (Val val@(MVal _)) = do
 
 -- Reduction for constants
 reduce _ (Val v) = do
-    return v
+    retVal v
 
 -- Reduction for variables
-reduce _ exp@(Var name) = do
-    rv <- findVar name
-    case rv of
-        (Just vexp) -> do
-            enterContext (cxtOfSigil $ head name) $ reduceExp vexp
+reduce env exp@(Var name) = do
+    v <- findVar env name
+    case v of
+        Just val -> reduce env (Val val)
         _ -> retError ("Undefined variable " ++ name) exp
 
 reduce _ (Statements stmts) = do
@@ -274,31 +263,33 @@ reduce _ (Statements stmts) = do
     
 -- Reduction for syntactic constructs
 reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
+    "block" -> do
+        let [body] = exps
+        enterBlock $ reduce env body
     "sub" -> do
         let [exp] = exps
         (VSub sub) <- enterEvalContext "Code" exp
         retVal $ VSub sub{ subPad = envLexical env }
-    "sym" -> do
-        mapM_ evalExp [ exp | Sym (Symbol _ _ exp) <- exps ]
-        retVal VUndef
     "mval" -> do
-        let [Var name, exp] = exps
-        val     <- enterEvalContext (cxtOfSigil $ head name) exp
-        retVal =<< newMVal val
+        let [exp] = exps
+        val     <- evalExp exp
+        val'    <- evalExp $ Val val
+        retVal =<< newMVal val'
     "if" -> doCond id 
     "unless" -> doCond not
     "for" -> do
         let [list, body] = exps
         vlist <- enterEvalContext "List" list
         vsub  <- enterEvalContext "Code" body
-        let arity = length (subParams $ vCast vsub)
+        VSub sub <- fromVal vsub
+        let arity = length (subParams sub)
             vals = concatMap vCast $ vCast vlist
             runBody [] = retVal VUndef
-            runBody (vs) = do
+            runBody vs = do
                 let (these, rest) = arity `splitAt` vs
-                doApply env (vCast vsub) [] $ map Val these
+                doApply env sub [] $ map Val these
                 runBody rest
-        runBody vals
+        enterLoop $ runBody vals
     "loop" -> do
         let [pre, cond, post, body] = exps
         evalExp pre
@@ -307,12 +298,37 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
             valPost <- evalExp post
             vbool   <- enterEvalContext "Bool" cond
             case valBody of
-                VError _ _ -> shiftT $ \_ -> return valBody
-                _ | VError _ _ <- valPost -> shiftT $ \_ -> return valPost
-                _ | not (vCast vbool) -> shiftT $ \_ -> return valBody
+                VError _ _ -> retVal valBody
+                _ | VError _ _ <- valPost -> retVal valPost
+                _ | not (vCast vbool) -> retVal valBody
                 _ -> runBody
-        val <- resetT runBody
-        retVal val
+        enterLoop runBody
+    "given" -> do
+        let [topic, body] = exps
+        vtopic <- enterEvalContext "Scalar" topic
+        enterGiven vtopic $ enterEvalContext "Code" body
+    "when" -> do
+        let [match, body] = exps
+        break <- evalVar "$?_BLOCK_EXIT"
+        match <- reduce env match
+        vtopic <- evalVar "$_"
+        topic <- reduce env $ Val vtopic
+        result <- op2Match topic match
+        let runBody = do
+            enterEvalContext "Code" body
+            VSub vbreak <- fromVal break
+            doApply env vbreak [] []
+        if (vCast result)
+            then enterWhen break runBody
+            else retVal VUndef
+    "default" -> do
+        let [body] = exps
+        break <- evalVar "$?_BLOCK_EXIT"
+        let runBody = do
+            enterEvalContext "Code" body
+            VSub vbreak <- fromVal break
+            doApply env vbreak [] []
+        enterWhen break runBody
     "while" -> doWhileUntil id
     "until" -> doWhileUntil not
     "=" -> do
@@ -326,7 +342,8 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
             [Var name, exp] -> do
                 val     <- evalVar name
                 val'    <- enterEvalContext (cxtOfSigil $ head name) exp
-                writeMVal val val'
+                val''   <- enterEvalContext (cxtOfSigil $ head name) (Val val')
+                writeMVal val val''
                 retVal val'
             [Syn "[]" [Var name, indexExp], exp] -> do
                 listMVal <- evalVar name
@@ -340,7 +357,14 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
                     list = concatMap vCast $ case listVal of
                         VUndef  -> [] -- autovivification
                         _       -> vCast listVal
-                    assignTo curr (index, val) = do
+                    assignTo curr (index, val)
+                        | index < 0
+                        , index' <- index + genericLength curr
+                        , index' >= 0
+                        = assignTo curr (index', val)
+                        | index < 0
+                        = retError "Modification of non-creatable array value attempted" (Val $ VInt index)
+                        | otherwise = do
                         pre <- preM
                         return $ pre ++ [val] ++ genericDrop (index + 1) curr
                         where
@@ -364,6 +388,7 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
                 writeMVal hashMVal $ VHash $ MkHash hash
                 retVal val'
             _ -> do
+                -- XXX LValue ??::
                 retError "Cannot modify constant item" (head exps)
     ":=" -> do
         let [Var name, exp] = exps
@@ -382,7 +407,7 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
         vals <- case exps of
             [Val v] | valType v == "Array" -> do
                 val <- enterEvalContext "List" $ head exps
-                fromValue val
+                fromVal val
             _ -> mapM (enterEvalContext "List") exps         
         cls  <- asks envClasses
         if isaType cls "Scalar" cxt          
@@ -422,8 +447,9 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
             else retVal $ VList slice
     "()" -> do
         let [subExp, Syn "invs" invs, Syn "args" args] = exps
-        sub     <- enterEvalContext "Code" subExp
-        apply (vCast sub) invs args
+        vsub <- enterEvalContext "Code" subExp
+        sub <- fromVal vsub
+        apply sub invs args
     "try" -> do
         val <- resetT $ evalExp (head exps)
         retEvalResult False val
@@ -431,6 +457,11 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
         val     <- enterEvalContext "List" exp
         -- ignore val
         retVal val
+    "rx" -> do
+        let [exp] = exps
+        val     <- enterEvalContext "Str" exp
+        str     <- fromVal val
+        retVal $ VRule $ mkRegex $ encodeUTF8 str
     syn | last syn == '=' -> do
         let [lhs, exp] = exps
             op = "&infix:" ++ init syn
@@ -439,6 +470,12 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
     where
     doSlice :: [Exp] -> [Val] -> [VInt] -> Maybe (Val, [VInt])
     doSlice errs vs (n:ns)
+        | n < 0
+        , n' <- n + genericLength vs
+        , n' >= 0
+        = doSlice errs vs (n':ns)
+        | n < 0
+        = Nothing
         | (v:_)         <- n `genericDrop` vs
         = Just (v, ns)
         | ((Val err):_) <- errs
@@ -454,23 +491,36 @@ reduce env@Env{ envContext = cxt } exp@(Syn name exps) = case name of
             else reduce env bodyElse
     -- XXX This treatment of while/until loops probably needs work
     doWhileUntil f = do
-        let [ cond, body] = exps
-        let ret val = shiftT $ \_ -> return val
-            runBody = do
+        let [cond, body] = exps
+        let runBody = do
             vbool <- enterEvalContext "Bool" cond
             case f $ vCast vbool of
                 True -> do
                     rv <- reduce env body
                     case rv of
-                        VError _ _  -> ret rv
+                        VError _ _  -> retVal rv
                         _           -> runBody
-                _ -> ret vbool
-        val <- resetT runBody
-        retVal val
+                _ -> retVal vbool
+        enterLoop runBody
 
 reduce env (App name [Syn "," invs] args) = reduce env (App name invs args)
 reduce env (App name invs [Syn "," args]) = reduce env (App name invs args)
 
+-- XXX absolutely evil bloody hack for "goto"
+reduce _ (App "&goto" (subExp:invs) args) = do
+    vsub <- enterEvalContext "Code" subExp
+    sub <- fromVal vsub
+    local callerEnv $ do
+        val <- apply sub invs args
+        shiftT $ \_ -> retVal val
+    where
+    callerEnv env = let caller = maybe env id (envCaller env) in
+        env{ envCaller = envCaller caller
+           , envContext = envContext caller
+           , envLValue = envLValue caller
+           }
+
+
 reduce Env{ envClasses = cls, envContext = cxt, envLexical = lex, envGlobal = glob } exp@(App name invs args) = do
     syms    <- liftIO $ readIORef glob
     subSyms <- mapM evalSym
@@ -478,8 +528,9 @@ reduce Env{ envClasses = cls, envContext = cxt, envLexical = lex, envGlobal = gl
         , let n = symName sym
         , (n ==) `any` [name, toGlobal name]
         ]
-    lens    <- mapM argSlurpLen (invs ++ args)
-    case findSub (sum lens) subSyms of
+    lens <- mapM argSlurpLen (invs ++ args)
+    sub <- findSub (sum lens) subSyms
+    case sub of
         Just sub    -> applySub subSyms sub invs args
         Nothing     -> retError ("No compatible subroutine found: " ++ name) exp
     where
@@ -517,12 +568,14 @@ reduce Env{ envClasses = cls, envContext = cxt, envLexical = lex, envGlobal = gl
             (App name' invs' args'):rest = invs
         syms    <- liftIO $ readIORef glob
         subSyms' <- mapM evalSym
+
             [ sym | sym <- lex ++ syms
             , let n = symName sym
             , (n ==) `any` [name', toGlobal name']
             ]
         lens'    <- mapM argSlurpLen (invs' ++ args')
-        case findSub (sum lens') subSyms' of
+        theSub <- findSub (sum lens') subSyms'
+        case theSub of
             Just sub'    -> applyChainSub subSyms' sub invs sub' invs' args' rest
             Nothing      -> apply sub{ subParams = (length invs) `replicate` p } invs [] -- XXX Wrong
             -- retError ("No compatible subroutine found: " ++ name') exp
@@ -535,24 +588,25 @@ reduce Env{ envClasses = cls, envContext = cxt, envLexical = lex, envGlobal = gl
         = apply sub{ subParams = (length invs) `replicate` p } invs [] -- XXX Wrong
         | otherwise
         = internalError "applyChainsub did not match a chain subroutine"
-    findSub slurpLen subSyms = case sort (subs slurpLen subSyms) of
-        ((_, sub):_)    -> Just sub
-        _               -> Nothing
-    subs slurpLen subSyms = [
-        ( (isGlobal, subT, isMulti sub, bound, distance)
-        , fromJust fun
-        )
-        | (n, val) <- subSyms
-        , let sub@(Sub{ subType = subT, subReturns = ret, subParams = prms }) = vCast val
-        , let isGlobal = '*' `elem` n
-        , let fun = arityMatch sub (length (invs ++ args)) slurpLen
-        , isJust fun
-        , deltaFromCxt ret /= 0
-        , let invocants = filter isInvocant prms
-        , let prms' = if null invocants then prms else invocants
-        , let distance = (deltaFromCxt ret : map (deltaFromScalar . paramContext) prms')
-        , let bound = either (const False) (const True) $ bindParams prms invs args
-        ]
+    findSub slurpLen subSyms = do
+        subs' <- subs slurpLen subSyms
+        return $ case sort subs' of
+            ((_, sub):_)    -> Just sub
+            _               -> Nothing
+    subs slurpLen subSyms = (liftM catMaybes) $ (`mapM` subSyms) $ \(n, val) -> do
+        sub@(Sub{ subType = subT, subReturns = ret, subParams = prms }) <- fromVal val
+        let isGlobal = '*' `elem` n
+        let fun = arityMatch sub (length (invs ++ args)) slurpLen
+        if not (isJust fun) then return Nothing else do
+        if not (deltaFromCxt ret /= 0) then return Nothing else do
+        let invocants = filter isInvocant prms
+        let prms' = if null invocants then prms else invocants
+        let distance = (deltaFromCxt ret : map (deltaFromScalar . paramContext) prms')
+        let bound = either (const False) (const True) $ bindParams prms invs args
+        return $ Just
+            ( (isGlobal, subT, isMulti sub, bound, distance)
+            , fromJust fun
+            )
     deltaFromCxt            = deltaType cls cxt
     deltaFromScalar ('*':x) = deltaFromScalar x
     deltaFromScalar x       = deltaType cls x "Scalar"
@@ -577,14 +631,13 @@ chainFun p1 f1 p2 f2 (v1:v2:vs) = do
 chainFun _ _ _ _ _ = internalError "chainFun: Not enough parameters in Val list"
 
 applyExp :: [ApplyArg] -> Exp -> Eval Val
-applyExp bound (Prim f)
-    = f [ argValue arg | arg <- bound, (argName arg !! 1) /= '_' ]
+applyExp bound (Prim f) =
+    f [ argValue arg | arg <- bound, (argName arg !! 1) /= '_' ]
 applyExp bound body = do
-    -- XXX - resetT here -- XXX - Wrong -- XXX - FIXME
     enterLex formal $ evalExp body
     where
     formal = filter (not . null . symName) $ map argNameValue bound
-    argNameValue (ApplyArg name val _) = Symbol SMy name (Val val)
+    argNameValue (ApplyArg name val _) = SymVal SMy name val
 
 apply :: VSub -> [Exp] -> [Exp] -> Eval Val
 apply sub invs args = do
@@ -599,9 +652,9 @@ doApply Env{ envClasses = cls } sub@Sub{ subParams = prms, subFun = fun, subType
         Left errMsg     -> retError errMsg (Val VUndef)
         Right bindings  -> do
             enterScope $ do
-                bound <- doBind bindings
+                (pad, bound) <- doBind [] bindings
                 -- trace (show bound) $ return ()
-                val <- local fixEnv $ do
+                val <- local fixEnv $ enterLex pad $ do
                     (`juncApply` bound) $ \realBound -> do
                         enterSub sub $ do
                             applyExp realBound fun
@@ -614,22 +667,23 @@ doApply Env{ envClasses = cls } sub@Sub{ subParams = prms, subFun = fun, subType
     fixEnv env
         | typ >= SubBlock = env
         | otherwise      = env{ envCaller = Just env }
-    doBind :: [(Param, Exp)] -> Eval [ApplyArg]
-    doBind [] = return []
-    doBind ((prm, exp):rest) = do
+    doBind :: Pad -> [(Param, Exp)] -> Eval (Pad, [ApplyArg])
+    doBind pad [] = return (pad, [])
+    doBind pad ((prm, exp):rest) = do
         -- trace ("<== " ++ (show (prm, exp))) $ return ()
-        (val, coll) <- case exp of
-            Parens exp  -> local fixEnv $ expToVal prm exp -- default
+        let name = paramName prm
+            cxt = cxtOfSigil $ head name
+        (val, coll) <- enterContext cxt $ case exp of
+            Parens exp  -> local fixEnv $ enterLex pad $ expToVal prm exp
             _           -> expToVal prm exp
         -- trace ("==> " ++ (show val)) $ return ()
-        let name = paramName prm
-            arg = ApplyArg name val coll
-        val' <- enterEvalContext (cxtOfSigil $ head name) (Val val)
-        restArgs <- enterLex [Symbol SMy name (Val val')] $ do
-            doBind rest
-        return (arg:restArgs)
-    expToVal Param{ isLValue = lv, isSlurpy = slurpy, paramContext = cxt } exp = do
-        val <- local (\e -> e{ envLValue = lv }) $ enterEvalContext cxt exp
+        let sym = SymVal SMy name val
+        (pad', restArgs) <- doBind (sym:pad) rest
+        return (pad', ApplyArg name val coll:restArgs)
+    expToVal Param{ isThunk = thunk, isLValue = lv, isSlurpy = slurpy, paramContext = cxt } exp = do
+        env <- ask -- freeze environment at this point for thunks
+        let eval = local (const env{ envLValue = lv }) $ enterEvalContext cxt exp
+        val <- if thunk then return (VThunk $ MkThunk eval) else eval
         return (val, (slurpy || isCollapsed cxt))
     isCollapsed cxt
         | isaType cls "Bool" cxt        = True
