@@ -23,6 +23,7 @@ import Parser
 import Help
 import Pretty
 import Posix
+import Prim
 
 main :: IO ()
 main = do
@@ -64,21 +65,26 @@ run []                          = do
         then do banner >> intro >> repLoop
         else run ["-"]
 
+-- convenience functions for GHCi
+eval = runProgramWith id (putStrLn . pretty) "<interactive>" []
+parse = doParse "-"
+dump = (doParseWith $ \exp _ -> print exp) "-"
+
 repLoop :: IO ()
 repLoop = do
-    command <- getCommand
-    case command of
-        CmdQuit             -> putStrLn "Leaving pugs."
-        CmdLoad fn          -> load fn
-        CmdEval trace prog  -> doEval trace [] prog >> repLoop
-        CmdParse prog       -> doParse "<interactive>" prog >> repLoop
-        CmdHelp             -> printInteractiveHelp >> repLoop
-        _                   -> internalError "unimplemented command"
+    env <- tabulaRasa >>= newIORef
+    modifyIORef env $ \e -> e{ envDebug = Nothing }
+    fix $ \loop -> do
+        command <- getCommand
+        case command of
+            CmdQuit           -> putStrLn "Leaving pugs."
+            CmdLoad fn        -> doLoad env fn >> loop
+            CmdRun opts prog  -> doRunSingle env opts prog >> loop
+            CmdParse prog     -> doParse "<interactive>" prog >> loop
+            CmdHelp           -> printInteractiveHelp >> loop
+            CmdReset          -> tabulaRasa >>= writeIORef env >> loop
 
-load _ = return ()
-
-parse = doParse "<interactive>"
-eval prog = doEval True [] prog
+tabulaRasa = prepareEnv "<interactive>" []
 
 doCheck = doParseWith $ \_ name -> do
     putStrLn $ name ++ " syntax OK"
@@ -96,18 +102,64 @@ doParseWith f name prog = do
         hPutStrLn stderr $ pretty err
         exitFailure
     f' exp = f exp name
-    
+
 
 doParse name prog = do
     env <- emptyEnv []
-    runRule env (putStrLn . pretty) ruleProgram name $ decodeUTF8 prog
+    case runRule env envBody ruleProgram name (decodeUTF8 prog) of
+        (Val err@(VError _ _)) -> putStrLn $ pretty err
+        exp -> putStrLn $ pretty exp
 
-doEval :: Bool -> [String] -> String -> IO ()
-doEval trace = do
-    runProgramWith f (putStrLn . pretty) "<interactive>"
+doLoad :: IORef Env -> String -> IO ()
+doLoad env fn = do
+    runImperatively env (evaluate exp)
+    return ()
     where
-    f | trace       = id
-      | otherwise   = \e -> e{ envDebug = Nothing }
+    exp = App "&require" [] [Val $ VStr fn]
+
+doRunSingle :: IORef Env -> RunOptions -> String -> IO ()
+doRunSingle menv opts prog = (`catch` handler) $ do
+    exp <- parse >>= makeProper
+    env <- theEnv
+    result <- runImperatively env (evaluate exp)
+    printer env result
+    where
+    parse = do
+        parseEnv <- emptyEnv []
+        runRule parseEnv (return . envBody) ruleProgram "<interactive>" (decodeUTF8 prog)
+    theEnv = do
+        ref <- if runOptSeparately opts
+                then tabulaRasa >>= newIORef
+                else return menv
+        debug <- if runOptDebug opts
+                then liftM Just (newIORef emptyFM)
+                else return Nothing
+        modifyIORef ref $ \e -> e{ envDebug = debug }
+        return ref
+    printer env = if runOptShowPretty opts
+        then \val -> do
+            final <- runImperatively env (fromVal' val)
+            putStrLn $ pretty final
+        else print
+    makeProper exp = case exp of
+        Val err@(VError _ _) -> fail $ pretty err
+        Statements stmts@((_,pos):_) | not (runOptSeparately opts) -> do
+            let withDump = stmts ++ [(Syn "dump" [], pos)]
+            return $ Statements withDump
+        _ | not (runOptSeparately opts) -> fail "Expected statements"
+        _ -> return exp
+    handler err = if not (isUserError err) then ioError err else do
+        putStrLn "Internal error while running expression:"
+        putStrLn $ ioeGetErrorString err
+
+runImperatively :: IORef Env -> Eval Val -> IO Val
+runImperatively menv eval = do
+    env <- readIORef menv
+    runEval env $ do
+        val <- eval
+        newEnv <- ask
+        liftIO $ writeIORef menv newEnv
+        return val
 
 doRun :: String -> [String] -> String -> IO ()
 doRun = do
@@ -130,10 +182,13 @@ runProgramWith fenv f name args prog = do
     val <- runEnv $ runRule (fenv env) id ruleProgram name $ decodeUTF8 prog
     f val
 
-runEnv env = do
+runEval :: Env -> Eval Val -> IO Val
+runEval env eval = do
     (`runReaderT` env) $ do
-        (`runContT` return) $ resetT $ do
-            evaluateMain (envBody env)
+        (`runContT` return) $
+            resetT eval
+
+runEnv env = runEval env $ evaluateMain (envBody env)
 
 runAST ast = do
     hSetBuffering stdout NoBuffering 
@@ -148,28 +203,45 @@ prepareEnv name args = do
     libs    <- getLibs environ
     progSV  <- newMVal $ VStr name
     endAV   <- newMVal $ VList []
+    matchAV <- newMVal $ VList []
     incAV   <- newMVal $ VList (map VStr libs)
     argsAV  <- newMVal $ VList (map VStr args)
     inGV    <- newMVal $ VHandle stdin
     outGV   <- newMVal $ VHandle stdout
     errGV   <- newMVal $ VHandle stderr
     errSV   <- newMVal $ VStr ""
+    let subExit = \x -> case x of
+            [x] -> op1 "exit" x
+            _   -> op1 "exit" VUndef
     emptyEnv
-        [ Symbol SGlobal "@*ARGS"       $ Val argsAV
-        , Symbol SGlobal "@*INC"        $ Val incAV
-        , Symbol SGlobal "$*PROGNAME"   $ Val progSV
-        , Symbol SGlobal "@*END"        $ Val endAV
-        , Symbol SGlobal "$*IN"         $ Val inGV
-        , Symbol SGlobal "$*OUT"        $ Val outGV
-        , Symbol SGlobal "$*ERR"        $ Val errGV
-        , Symbol SGlobal "$!"           $ Val errSV
-        , Symbol SGlobal "%*ENV" (Val . VHash . MkHash $ envFM)
+        [ SymVal SGlobal "@*ARGS"       argsAV
+        , SymVal SGlobal "@*INC"        incAV
+        , SymVal SGlobal "$*PROGNAME"   progSV
+        , SymVal SGlobal "@*END"        endAV
+        , SymVal SGlobal "$*IN"         inGV
+        , SymVal SGlobal "$*OUT"        outGV
+        , SymVal SGlobal "$*ERR"        errGV
+        , SymVal SGlobal "$!"           errSV
+        , SymVal SGlobal "$/"           matchAV
+        , SymVal SGlobal "%*ENV" (VHash . MkHash $ envFM)
         -- XXX What would this even do?
-        -- , Symbol SGlobal "%=POD"        (Val . VHash . MkHash $ emptyFM) 
-        , Symbol SGlobal "@=POD"        (Val . VArray . MkArray $ [])
-        , Symbol SGlobal "$=POD"        (Val . VStr $ "")
-        , Symbol SGlobal "$?OS"         (Val . VStr $ config_osname)
+        -- , SymVal SGlobal "%=POD"        (Val . VHash . MkHash $ emptyFM) 
+        , SymVal SGlobal "@=POD"        (VArray . MkArray $ [])
+        , SymVal SGlobal "$=POD"        (VStr "")
+        , SymVal SGlobal "$?OS"         (VStr config_osname)
+        , SymVal SGlobal "$?_BLOCK_EXIT" $ VSub $ Sub
+            { isMulti = False
+            , subName = "$?_BLOCK_EXIT"
+            , subType = SubPrim
+            , subPad = []
+            , subAssoc = "pre"
+            , subParams = []
+            , subReturns = "Void"
+            , subFun = Prim subExit
+            }
         ]
+       
+
 
 getLibs :: [(String, String)] -> IO [String]
 getLibs environ = return $ filter (not . null) libs
