@@ -67,6 +67,51 @@ enterScope f = do
     return rv
     -}
 -}
+
+enterGiven topic action = enterLex [SymVal SMy "$_" topic] action
+
+enterWhen break action = callCC $ \esc -> do
+    enterLex [SymVal SMy "&continue" $ continueSub esc,
+              SymVal SMy "&break" break] action
+    where
+    continueSub esc = VSub $ Sub
+        { isMulti = False
+        , subName = "continue"
+        , subType = SubPrim
+        , subPad = []
+        , subAssoc = "pre"
+        , subParams = []
+        , subReturns = "Void"
+        , subFun = Prim (const $ esc VUndef)
+        }
+
+enterLoop action = callCC $ \esc -> do
+    enterLex [SymVal SMy "&last" $ lastSub esc] action
+    where
+    lastSub esc = VSub $ Sub
+        { isMulti = False
+        , subName = "last"
+        , subType = SubPrim
+        , subPad = []
+        , subAssoc = "pre"
+        , subParams = []
+        , subReturns = "Void"
+        , subFun = Prim (const $ esc VUndef)
+        }
+
+enterBlock action = callCC $ \esc -> do
+    enterLex [SymVal SMy "$?_BLOCK_EXIT" $ escSub esc] action
+    where
+    escSub esc = VSub $ Sub
+        { isMulti = False
+        , subName = "$?_BLOCK_EXIT"
+        , subType = SubPrim
+        , subPad = []
+        , subAssoc = "pre"
+        , subParams = []
+        , subReturns = "Void"
+        , subFun = Prim (const $ esc VUndef)
+        }
   
 enterSub sub@Sub{ subType = typ } action
     | typ >= SubPrim = action -- primitives just happen
@@ -81,11 +126,11 @@ enterSub sub@Sub{ subType = typ } action
     doReturn _   = internalError "enterSub: doReturn list length /= 1"
     doCC cc [v] = cc v
     doCC _  _   = internalError "enterSub: doCC list length /= 1"
-    subRec = [ Symbol SMy "&?SUB" (Val $ VSub sub)
-             , Symbol SMy "$?SUBNAME" (Val $ VStr $ subName sub)]
-    blockRec = Symbol SMy "&?BLOCK" (Val $ VSub sub)
-    ret cxt = Symbol SMy "&return" (Val $ VSub $ retSub cxt)
-    callerCC cc cxt = Symbol SMy "&?CALLER_CONTINUATION" (Val $ VSub $ ccSub cc cxt)
+    subRec = [ SymVal SMy "&?SUB" (VSub sub)
+             , SymVal SMy "$?SUBNAME" (VStr $ subName sub)]
+    blockRec = SymVal SMy "&?BLOCK" (VSub sub)
+    ret cxt = SymVal SMy "&return" (VSub $ retSub cxt)
+    callerCC cc cxt = SymVal SMy "&?CALLER_CONTINUATION" (VSub $ ccSub cc cxt)
     fixEnv cc pad cxt env
         | typ >= SubBlock = env{ envLexical = (blockRec:subPad sub) ++ pad }
         | otherwise      = env{ envLexical = subRec ++ (ret cxt:callerCC cc cxt:subPad sub) }
@@ -101,6 +146,7 @@ enterSub sub@Sub{ subType = typ } action
             , isOptional = False
             , isNamed = False
             , isLValue = False
+            , isThunk = False
             , paramName = "@?0"
             , paramContext = cxt
             , paramDefault = Val VUndef
@@ -120,6 +166,7 @@ enterSub sub@Sub{ subType = typ } action
             , isOptional = False
             , isNamed = False
             , isLValue = False
+            , isThunk = False
             , paramName = "@?0"
             , paramContext = cxt
             , paramDefault = Val VUndef
@@ -140,7 +187,7 @@ innerSub = Sub
     { isMulti       = False
     , subName       = "inner"
     , subType       = SubRoutine
-    , subPad        = [Symbol SMy "$inner" (Val VUndef)]
+    , subPad        = [SymVal SMy "$inner" VUndef]
     , subAssoc      = "left"
     , subParams     = []
     , subReturns    = "List"
@@ -151,7 +198,7 @@ sub3Sub = Sub
     { isMulti       = False
     , subName       = "sub3"
     , subType       = SubRoutine
-    , subPad        = [Symbol SMy "$inner" (Val VUndef)]
+    , subPad        = [SymVal SMy "$inner" VUndef]
     , subAssoc      = "left"
     , subParams     = []
     , subReturns    = "List"
@@ -170,7 +217,7 @@ dumpLex label = do
 blah :: Eval Val
 blah = do
     dumpLex ">init"
-    rv <- enterLex [Symbol SMy "$x" $ Val $ VInt 1] $ do
+    rv <- enterLex [SymVal SMy "$x" (VInt 1)] $ do
         dumpLex ">lex"
         -- rv <- enterScope outer
         rv <- outer
@@ -180,7 +227,7 @@ blah = do
     return rv
 
 outer :: Eval Val
-outer = enterLex [Symbol SMy "$outer" $ Val $ VInt 2] $ do
+outer = enterLex [SymVal SMy "$outer" (VInt 2)] $ do
     dumpLex ">outer"
     -- enterSub innerSub
     dumpLex "<outer"
